use std::{process::Stdio, sync::Arc};

use anyhow::anyhow;
use itertools::Itertools;
use tokio::{
    net::TcpStream,
    process::{Child, Command},
    sync::Mutex,
    time::{sleep, Duration},
};

use crate::{config::ModelSpec, error::AppError, state::AppState, util::parse_size};

#[derive(Clone, Debug)]
pub struct InferenceProcess {
    pub spec:    ModelSpec,
    pub process: Arc<Mutex<Child>>,
}

impl InferenceProcess {
    /// Retrieve a running process from state or spawn a new one.
    pub async fn get_or_spawn(
        spec: &ModelSpec,
        state: AppState,
    ) -> Result<InferenceProcess, AppError> {
        let required_ram = parse_size(&spec.ram_usage).expect("Invalid ram_usage in model spec");
        let required_vram = parse_size(&spec.vram_usage).expect("Invalid vram_usage in model spec");

        {
            let state_guard = state.lock().await;
            if let Some(proc) = state_guard.processes.get(&spec.port) {
                return Ok(proc.clone());
            }
        }

        let mut state_guard = state.lock().await;
        tracing::info!(msg = "App state before spawn", ?state_guard);

        // If not enough resources, stop some running processes.
        if (state_guard.used_ram + required_ram > state_guard.total_ram)
            || (state_guard.used_vram + required_vram > state_guard.total_vram)
        {
            let mut to_remove = Vec::new();
            let processes_by_usage = state_guard
                .processes
                .iter()
                .sorted_by(|(_, a), (_, b)| {
                    parse_size(&b.spec.vram_usage)
                        .unwrap_or(0)
                        .cmp(&parse_size(&a.spec.vram_usage).unwrap_or(0))
                })
                .map(|(port, proc)| (*port, proc.clone()))
                .collect::<Vec<_>>();

            for (port, proc) in processes_by_usage.iter() {
                tracing::info!("Stopping process on port {}", port);
                let mut lock = proc.process.lock().await;
                lock.kill().await.ok();
                to_remove.push(*port);
                state_guard.used_ram = state_guard
                    .used_ram
                    .saturating_sub(parse_size(&proc.spec.ram_usage).unwrap_or(0));
                state_guard.used_vram = state_guard
                    .used_vram
                    .saturating_sub(parse_size(&proc.spec.vram_usage).unwrap_or(0));
                if state_guard.used_ram + required_ram <= state_guard.total_ram
                    && state_guard.used_vram + required_vram <= state_guard.total_vram
                {
                    tracing::info!("Freed enough resources");
                    break;
                }
            }
            for port in to_remove {
                state_guard.processes.remove(&port);
            }
        } else {
            tracing::info!("Sufficient resources available");
        }

        let proc = Self::spawn(spec).await?;
        state_guard.used_ram += required_ram;
        state_guard.used_vram += required_vram;
        state_guard.processes.insert(spec.port, proc.clone());

        sleep(Duration::from_millis(250)).await;
        proc.wait_until_ready().await?;
        sleep(Duration::from_millis(250)).await;
        Ok(proc)
    }

    /// Spawn a new inference process.
    pub async fn spawn(spec: &ModelSpec) -> Result<InferenceProcess, AppError> {
        let args = spec
            .args
            .iter()
            .flat_map(|(k, v)| {
                if v == "true" {
                    vec![format!("--{}", k)]
                } else {
                    vec![format!("--{}", k), v.clone()]
                }
            })
            .collect::<Vec<_>>();

        let internal_port = spec.internal_port.expect("Internal port must be set");

        let mut cmd = Command::new("llama-server");
        cmd.kill_on_drop(true)
            .envs(spec.env.clone())
            .args(&args)
            .arg("--port")
            .arg(format!("{}", internal_port))
            .stdout(Stdio::null())
            .stderr(Stdio::null());

        tracing::info!("Starting llama-server with command: {:?}", cmd);
        let child = cmd.spawn().expect("Failed to start llama-server");

        Ok(InferenceProcess {
            spec:    spec.clone(),
            process: Arc::new(Mutex::new(child)),
        })
    }

    pub async fn wait_until_ready(&self) -> Result<(), anyhow::Error> {
        async fn check_running(proc: &InferenceProcess) -> Result<(), AppError> {
            let mut lock = proc.process.clone().lock_owned().await;
            match lock.try_wait()? {
                Some(exit_status) => {
                    tracing::error!("Inference process exited: {:?}", exit_status);
                    Err(AppError::Unknown(anyhow!("Inference process exited")))
                }
                None => Ok(()),
            }
        }

        async fn wait_for_port(port: u16) -> Result<(), anyhow::Error> {
            for _ in 0..10 {
                if TcpStream::connect(("127.0.0.1", port)).await.is_ok() {
                    return Ok(());
                }
                sleep(Duration::from_millis(500)).await;
            }
            Err(anyhow!("Timeout waiting for port"))
        }

        check_running(self).await?;
        wait_for_port(self.spec.internal_port.expect("Internal port must be set")).await?;
        Ok(())
    }
}
