pub mod config;
pub mod error;
pub mod inference_process;
pub mod logging;
pub mod proxy;
pub mod state;
pub mod util;

use std::net::SocketAddr;

use axum::{routing::any, Router};
use config::{AppConfig, ModelSpec};
use state::AppState;
use tower_http::trace::{
    DefaultMakeSpan,
    DefaultOnEos,
    DefaultOnFailure,
    DefaultOnRequest,
    DefaultOnResponse,
    TraceLayer,
};
use tracing::Level;

/// Creates an Axum application to handle inference requests for a specific model.
pub fn axum_router(spec: &ModelSpec, state: AppState) -> Router {
    Router::new()
        .route(
            "/",
            any({
                let state = state.clone();
                let spec = spec.clone();
                move |req| proxy::handle_request(req, spec.clone(), state.clone())
            }),
        )
        .route(
            "/*path",
            any({
                let state = state.clone();
                let spec = spec.clone();
                move |req| proxy::handle_request(req, spec.clone(), state.clone())
            }),
        )
        .layer(
            TraceLayer::new_for_http()
                .make_span_with(DefaultMakeSpan::new().include_headers(true))
                .on_request(DefaultOnRequest::new().level(Level::DEBUG))
                .on_response(DefaultOnResponse::new().level(Level::TRACE))
                .on_eos(DefaultOnEos::new().level(Level::DEBUG))
                .on_failure(DefaultOnFailure::new().level(Level::ERROR)),
        )
}

/// Starts an inference server for each model defined in the config.
pub async fn start_server(config: AppConfig) -> anyhow::Result<()> {
    let state = AppState::from_config(config.clone());

    let mut handles = Vec::new();
    for spec in config.model_specs {
        let state = state.clone();
        let spec = spec.clone();

        let handle: tokio::task::JoinHandle<anyhow::Result<()>> = tokio::spawn(async move {
            let app = axum_router(&spec, state);
            let addr = SocketAddr::from(([0, 0, 0, 0], spec.port));
            tracing::info!(msg = "Listening", ?spec);
            let listener = tokio::net::TcpListener::bind(&addr).await?;
            axum::serve(listener, app.into_make_service()).await?;
            Ok(())
        });
        handles.push(handle);
    }

    let _ = futures::future::try_join_all(handles).await?;
    Ok(())
}
